{Detect Sound Blaster port, IRQ and DMA settings}
{Copyright (C) 1998-99 by Andrew Below (bel@obninsk.ru)}

Program DetectSB;

{$DEFINE SHOWINFO}

Uses DOS, CRT;

Var
 SB_BASE: Word;
 SB_IRQ: Byte;
 SB_DMA: Byte;

Var
 DSP_Reset: Word;
 DSP_ReadData: Word;
 DSP_WriteData: Word;
 DSP_WriteStatus: Word;
 DSP_DataStatus: Word;
 MIX_RegPort: Word;
 MIX_RegData: Word;

Const
 DSPWriteReadCount  = 100000;
 DSPResetCount      = 20;
 IRQTestWaitCount   = 5;
 DMATestWaitCount   = 10;
 DMATestBufSize     = 2048;
 DMATestSampleRate  = 11025;


procedure dev_irq_enable(i: word); forward;
procedure dev_irq_disable(i: word); forward;
function  dev_irq_state(i: word): wordbool; forward;
function  dev_irq_setvector(i: word; vec: pointer): pointer; forward;
function  dev_irq_getvector(i: word): pointer; forward;
procedure dev_dma_init(channel, mode: byte; buf: pointer; size: word); forward;
procedure dev_dma_reset(channel: byte); forward;
function  dev_dma_getcount(channel: byte): word; forward;


const
 dev_irq_bitmask: array[0..7] of byte = ($FE, $FD, $FB, $F7, $EF, $DF, $BF, $7F);


{**************************************}
{ 뢠  㪠 IRQ}
{**************************************}
procedure dev_irq_enable(i: word);
begin
 asm cli end;
 if i < 8 then
 port[$21] := port[$21] and (not (1 shl i)) else
 begin
  port[$21] := port[$21] and (not (1 shl 2));
  port[$A1] := port[$A1] and (not (1 shl (i-8)));
 end;
 asm sti end;
end;


{**************************************}
{ 뢠  㪠 IRQ}
{**************************************}
procedure dev_irq_disable(i: word);
begin
 asm cli end;
 if i < 8 then
 port[$21] := port[$21] or (1 shl i) else
 port[$A1] := port[$A1] or (1 shl (i-8));
 asm sti end;
end;


{****************************************************}
{  㪠 IRQ (ࠧ襭/饭)}
{****************************************************}
function dev_irq_state(i: word): wordbool;
begin
 if i < 8 then dev_irq_state := (port[$21] and (1 shl i)) = 0 else
 dev_irq_state := ((port[$21] and 4) = 0) and ((port[$A1] and (1 shl (i-8))) = 0);
end;


{**************************************************}
{⠭  IRQ (頥  )}
{**************************************************}
function dev_irq_setvector(i: word; vec: pointer): pointer;
var
 p: pointer;
 irqnum: byte;
begin
 if i < 8 then irqnum := i+$08 else irqnum := i+$68;
 getintvec(irqnum, p);
 setintvec(irqnum, vec);
 dev_irq_setvector := p;
end;


{**********************************}
{頥   㪠 IRQ}
{**********************************}
function dev_irq_getvector(i: word): pointer;
var
 p: pointer;
 irqnum: byte;
begin
 if i<8 then irqnum := i+$08 else irqnum := i+$68;
 getintvec(irqnum, p);
 dev_irq_getvector := p;
end;


const
 dev_dma_lpage: array[0..7] of word = ($87, $83, $81, $82, $88, $8B, $89, $8A);
 dev_dma_hpage: array[0..7] of word = ($487,$483,$481,$482,$00,$48B,$489,$48A);
 dev_dma_addr:  array[0..7] of word = ($00, $02, $04, $06, $C0, $C4, $C8, $CC);
 dev_dma_count: array[0..7] of word = ($01, $03, $05, $07, $C2, $C6, $CA, $CE);

const
 dev_dma_singleplayback   = $48;
 dev_dma_autoinitplayback = $58;
 dev_dma_singlerecord     = $44;
 dev_dma_autoinitrecord   = $54;


{***********************************************}
{樠஢ DMA-  뫪 }
{***********************************************}
procedure dev_dma_init(channel, mode: byte; buf: pointer; size: word);
var
 offset: word;
 page: byte;
begin
 offset := (seg(buf^) shl 4)+ofs(buf^);
 page := (seg(buf^)+ofs(buf^) shr 4) shr 12;

 if channel>3 then
 begin
  port[$D4] := channel;
  port[$D8] := 0;
  port[$D6] := mode+channel-4;
 end else
 begin
  port[$0A] := channel+4;
  port[$0C] := 0;
  port[$0B] := mode+channel;
 end;
 port[dev_dma_addr[channel]] := lo(offset);
 port[dev_dma_addr[channel]] := hi(offset);
 port[dev_dma_lpage[channel]] := page;
 port[dev_dma_count[channel]] := lo(size-1);
 port[dev_dma_count[channel]] := hi(size-1);
 if channel>3 then port[$D4] := Channel-4
 else port[$0A] := channel;
end;


{******************}
{ DMA-}
{******************}
procedure dev_dma_reset(channel: byte);
begin
 if channel>3 then port[$D4] := channel
 else port[$0A] := channel+4;
end;


{****************************************************}
{頥  뫪  㪠  DMA}
{****************************************************}
function dev_dma_getcount(channel: byte): word;
var
 w: word;
begin
 if channel>3 then port[$D8] := 0
 else port[$0C] := 0;
 w := port[dev_dma_count[channel]];
 w := word(port[dev_dma_count[channel]]) shl 8;
 w := w+1;
 if channel>3 then w := w shl 1;
 dev_dma_getcount := w;
end;


Function Word2Hex(W: Word): String; Assembler;
Const
 Hex: Array[0..$F] Of Char = '0123456789ABCDEF';
Asm
 Les   DI, @Result
 Mov   AL, 5
 StoSB
 Mov   BX, W
 Mov   DX, BX
 Mov   BL, BH
 Xor   BH, BH
 Mov   CL, 4
 Shr   BX, CL
 Mov   AL, Byte Ptr Hex[BX]
 Mov   BL, DH
 And   BX, $0F
 Mov   AH, Byte Ptr Hex[BX]
 StoSW
 Mov   BL, DL
 Xor   BH, BH
 Mov   CL, 4
 Shr   BX, CL
 Mov   AL, Byte Ptr Hex[BX]
 Mov   BL, DL
 And   BX, $0F
 Mov   AH, Byte Ptr Hex[BX]
 StoSW
 Mov   AL, 'h'
 StoSB
End;


{************}
{Reset SB DSP}
{************}
Function sbResetDSP: Boolean;
Var
 StartTimer, Ticks: LongInt;
 ResetOK: Boolean;
 OKByte: Byte;
Begin
 ResetOK := False;
 StartTimer := MemL[$40:$6C];
 {Reset DSP}
 While (Ticks < DSPResetCount) And (Not ResetOK) Do
 Begin
  Port[DSP_Reset] := 1; Delay(3);
  Port[DSP_Reset] := 0; Delay(3);
  If (Port[DSP_DataStatus] And $80 = $80) Then OKByte := Port[DSP_ReadData];
  If OKByte = $AA Then ResetOK := True;
  Ticks := MemL[$40:$6C] - StartTimer;
 End;
 sbResetDSP := ResetOK;
End;


{************************}
{Write single byte to DSP}
{************************}
Procedure sbWriteDSP(Value: Byte);
Var
 Count: LongInt;
Begin
 Count := DSPWriteReadCount;
 While (Port[DSP_WriteStatus] And $80 <> 0) And (Count <> 0) Do Dec(Count);
 Port[DSP_WriteData] := Value;
End;


{*************************}
{Read single byte from DSP}
{*************************}
Function sbReadDSP: Byte;
Var
 Count: LongInt;
Begin
 Count := DSPWriteReadCount;
 While (Port[DSP_DataStatus] And $80 = 0) And (Count <> 0) Do Dec(Count);
 sbReadDSP := Port[DSP_ReadData];
End;


{***********************************}
{Get DSP version as formatted string}
{***********************************}
Function sbGetDSPVersion: String;
Var
 MajorByte, MinorByte: Byte;
 MajorStr, MinorStr: String;
Begin
 sbWriteDSP($E1);
 MajorByte := sbReadDSP;
 Str(MajorByte, MajorStr);
 MinorByte := sbReadDSP;
 Str(MinorByte, MinorStr);
 If MinorByte < 10 Then MinorStr := '0' + MinorStr;
 sbGetDSPVersion := MajorStr + '.' + MinorStr;
End;


{**********************}
{Detect SB base address}
{**********************}
Function sbDetectBase(Var SB_BASE: Word): Byte;
Var
 StartTimer, Ticks: LongInt;
 BaseDetected: Boolean;
 B: Byte;
Begin
 BaseDetected := False;
 SB_BASE := $200;
 While (SB_BASE <= $280) Do
 Begin
  {$IFDEF SHOWINFO}
  WriteLn('Checking base address ', Word2Hex(SB_BASE), '...');
  {$ENDIF}
  {Calculate DSP ports}
  DSP_Reset := SB_BASE + $06;
  DSP_ReadData := SB_BASE + $0A;
  DSP_WriteData := SB_BASE + $0C;
  DSP_WriteStatus := SB_BASE + $0C;
  DSP_DataStatus := SB_BASE + $0E;
  MIX_RegPort := SB_BASE + $04;
  MIX_RegData := SB_BASE + $05;
  If (Port[SB_BASE] <> $FF) Then
  Begin
   StartTimer := MemL[$40:$6C];
   Ticks := 0;
   While (Ticks < DSPResetCount) And (Not BaseDetected) Do
   Begin
    Port[DSP_Reset] := 1; Delay(3);
    Port[DSP_Reset] := 0; Delay(3);
    If sbReadDSP = $AA Then BaseDetected := True;
    Ticks := MemL[$40:$6C] - StartTimer;
   End;
   If BaseDetected Then Break;
  End;
 Inc(SB_BASE, $10);
 End;
 If (SB_BASE > $280) Or (Not BaseDetected) Then
 Begin SB_BASE := 0; sbDetectBase := 1; Exit; End;
 sbDetectBase := 0;
 {  ⪨ ⠢ SB  ଠ  DSP...}
 For Ticks := 1 To 10000 Do
 Begin
  Port[DSP_WriteData] := $E1;
  B := Port[DSP_ReadData];
  B := Port[DSP_ReadData];
  Port[DSP_DataStatus] := $80;
  Port[DSP_WriteStatus] := $80;
 End;
End;


Var
 sbOldIRQHandler: Pointer;
 sbIRQCalled: Boolean;
 sbTempIRQNum: Byte;
 sbTempDMANum: Byte;


{*******************}
{Test SB IRQ handler}
{*******************}
Procedure sbIRQHandler; Interrupt;
Var
 Info: Byte;
Begin
 Asm Cli End;
 sbIRQCalled := True;
 Info := Port[DSP_DataStatus];
 If sbTempIRQNum > 7 Then Port[$A0] := $20;
 Port[$20] := $20;
 Asm Sti End;
End;


{******************}
{Detect SB IRQ line}
{******************}
Function sbDetectIRQ(Var SB_IRQ: Byte): Byte;
Const
 TotalIRQsToTest = 4;
 IRQs: Array[1..TotalIRQsToTest] Of Byte = (5, 7, 9, 10);
Var
 StartTimer, Ticks: LongInt;
 CurrentIRQ: Byte;
Begin
 For CurrentIRQ := 1 To TotalIRQsToTest Do
 Begin
  sbTempIRQNum := IRQs[CurrentIRQ];
  {$IFDEF SHOWINFO}
  WriteLn('Checking IRQ', sbTempIRQNum, '...');
  {$ENDIF}
  sbOldIRQHandler := dev_irq_setvector(sbTempIRQNum, @sbIRQHandler);
  dev_irq_enable(sbTempIRQNum);
  sbIRQCalled := False;
  sbWriteDSP($F2);
  StartTimer := MemL[$40:$6C];
  Ticks := 0;
  While Ticks < IRQTestWaitCount Do Ticks := MemL[$40:$6C] - StartTimer;
  dev_irq_disable(sbTempIRQNum);
  dev_irq_setvector(sbTempIRQNum, sbOldIRQHandler);
  If sbIRQCalled Then
  Begin
   SB_IRQ := sbTempIRQNum;
   sbDetectIRQ := 0;
   Exit;
  End;
 End;
 sbDetectIRQ := 1;
End;


{***************}
{Turn speaker on}
{***************}
Procedure sbSpeakerOn;
Begin
 sbWriteDSP($D1);
End;


{****************}
{Turn speaker off}
{****************}
Procedure sbSpeakerOff;
Begin
 sbWriteDSP($D3);
End;


{****************************************}
{Start single-cycle DMA playback transfer}
{****************************************}
Procedure sbTestPlayback(DMACh: Byte; Sound: Pointer; Size, SampleRate: Word);
Begin
 sbSpeakerOn;
 dev_dma_init(DMACh, $48, Sound, Size);
 sbWriteDSP($40);
 sbWriteDSP(256-(1000000 Div LongInt(SampleRate)));
 sbWriteDSP($14);
 sbWriteDSP(Lo(Size-1));
 sbWriteDSP(Hi(Size-1));
 sbIRQCalled := False;
End;


{*********************}
{Detect SB DMA channel}
{*********************}
Function sbDetectDMA(Var SB_DMA: Byte): Byte;
Const
 TotalDMAsToTest = 3;
 DMAs: Array[1..TotalDMAsToTest] Of Byte = (0, 1, 3);
Var
 TestDMABuffer, ExDMABuffer: Pointer;
 StartTimer, Ticks: LongInt;
 CurrentDMA: Byte;
 P, DMAPos: Word;
Begin
 {Allocate memory for DMA buffers}
 GetMem(ExDMABuffer, 16);
 GetMem(TestDMABuffer, DMATestBufSize);
 P := 16;
 While (Seg(TestDMABuffer^) Mod 4096) > (4096-(DMATestBufSize * 2 Div 16)) Do
 Begin
  FreeMem(TestDMABuffer, DMATestBufSize);
  FreeMem(ExDMABuffer, P);
  Inc(P, 16);
  If P > 65520 Then
  Begin WriteLn('DMA: cannot allocate memory.'); Halt; End;
  GetMem(ExDMABuffer, P);
  GetMem(TestDMABuffer, DMATestBufSize);
 End;
 FillChar(TestDMABuffer^, DMATestBufSize, $80);
 FreeMem(ExDMABuffer, P);

 For CurrentDMA := 1 To TotalDMAsToTest Do
 Begin
  sbTempDMANum := DMAs[CurrentDMA];
  {$IFDEF SHOWINFO}
  WriteLn('Checking DMA', sbTempDMANum, '...');
  {$ENDIF}
  sbResetDSP;
  dev_dma_reset(sbTempDMANum);
  sbTestPlayback(sbTempDMANum, TestDMABuffer, DMATestBufSize, DMATestSampleRate);
  StartTimer := MemL[$40:$6C];
  Ticks := 0;
  While Ticks < DMATestWaitCount Do Ticks := MemL[$40:$6C] - StartTimer;
  DMAPos := dev_dma_getcount(sbTempDMANum);
  sbResetDSP;
  dev_dma_reset(sbTempDMANum);
  If DMAPos > DMATestBufSize Then
  Begin
   sbDetectDMA := 0;
   SB_DMA := sbTempDMANum;
   Exit;
  End;
 End;
 sbDetectDMA := 1;
End;


Const
 clrOK       = 3;
 clrMsg      = 2;
 clrDebugMsg = 7;
 clrError    = 4;

Begin
 TextAttr := 7; ClrScr;

 TextAttr := clrMsg;
 WriteLn('1. Detecting SB base address...');
 TextAttr := clrDebugMsg;
 If sbDetectBase(SB_BASE) = 0 Then
 Begin
  TextAttr := clrOK;
  WriteLn('SB detected at base address ', Word2Hex(SB_BASE), '.');
 End Else
 Begin
  TextAttr := clrError;
  WriteLn('SB not detected.');
 End;
 If SB_BASE = 0 Then Halt;
 WriteLn('SB DSP version is ', sbGetDSPVersion, '.');
 WriteLn;

 TextAttr := clrMsg;
 WriteLn('2. Detecting SB IRQ line...');
 TextAttr := clrDebugMsg;
 If sbDetectIRQ(SB_IRQ) = 0 Then
 Begin
  TextAttr := clrOK;
  WriteLn('SB detected at IRQ', SB_IRQ, '.')
 End Else
 Begin
  TextAttr := clrError;
  WriteLn('Cannot detect SB IRQ line.');
 End;
 WriteLn;

 TextAttr := clrMsg;
 WriteLn('3. Detecting SB DMA channel...');
 TextAttr := clrDebugMsg;
 If sbDetectDMA(SB_DMA) = 0 Then
 Begin
  TextAttr := clrOK;
  WriteLn('SB detected at DMA', SB_DMA, '.')
 End Else
 Begin
  TextAttr := clrError;
  WriteLn('Cannot detect SB DMA channel.');
 End;
End.